# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
""" Copyright (c) 2000-2003 LOGILAB S.A. (Paris, FRANCE).
 http://www.logilab.fr/ -- mailto:contact@logilab.fr

Python code format's checker

Some the process_token method is based from The Tab Nanny std module.
"""

__revision__ = "$Id: format.py,v 1.3 2003/06/03 09:11:57 syt Exp $"

import tokenize
if not hasattr(tokenize, 'NL'):
    raise ValueError("tokenize.NL doesn't exist -- tokenize module too old")
import re

from logilab.pylint.interfaces import IRawChecker
from logilab.pylint.checkers import BaseRawChecker, CheckerHandler


OPERATORS = '!=|<=|==|>=|<|>|=|\+=|-=|\*=|/='
OP_RGX = '[^<>=+\-*/!]+((?<!\s)(%s)|((%s)(?!\s|=|>))).*' % (OPERATORS,
                                                            OPERATORS)

SC_RGX = '(?:\w|\s)+:\s*\w+'
C_RGX = '(?:\w|\s)+,[^\s]'

BAD_CONSTRUCT_RGX = re.compile('(?:%s)|(:?%s)|(:?%s)' % (SC_RGX, C_RGX, OP_RGX))


class IndentError(Exception):
    """ expression raisen on ambiguous indentation """
 
class FormatChecker(BaseRawChecker, CheckerHandler):
    """checks for :                                                            
    * unauthorized constructions                                               
    * strict indentation                                                       
    * lines'length                                                             
    """
    
    __implements__ = IRawChecker

    # configuration section name
    name = 'format'
    # configuration options
    # for available dict keys/values see the optik parser 'add_option' method
    options = (('max-line-length',
                {'default' : 80, 'type' : "int", 'metavar' : '<int>',
                 'help' : 'Maximum number of characters on a single line.'}),
               ('max-module-lines',
                {'default' : 1000, 'type' : 'int', 'metavar' : '<int>',
                 'help': 'Maximum number of lines in a module'}
                ),
               ('indent-string',
                {'default' : '    ', 'type' : "string", 'metavar' : '<string>',
                 'help' : 'String used as indentation unit. This is usually \
"    " (4 spaces) or "\\t" (1 tab).'}),
               ('bad-construction-rgx',
                {'default' : BAD_CONSTRUCT_RGX,
                 'type' : "regexp", 'metavar' : '<regexp>',
                 'help' : 'Regular expression which should match bad \
constructions.'}),
               )
    
    def process_tokens(self, tokens):
        """ process tokens and search for :
    
         _ non strict indentation (i.e. not always using the <indent> parameter as
           indent unit)
         _ too long lines (i.e. longer than <max_chars>)
         _ optionally bad construct (if given, bad_construct must be a compiled
           regular expression).
        """
        indent = tokenize.INDENT
        dedent = tokenize.DEDENT
        newline = tokenize.NEWLINE
        junk = tokenize.COMMENT, tokenize.NL
        indent_string = self.config.indent_string
        # add 1 for the trailing '\n'
        max_chars = self.config.max_line_length + 1 
        bad_construct = self.config.bad_construction_rgx
        indents = [0]
        check_equal = 0
        line_num = 0
        long_lines = {}
        for (type, token, start, end, line) in tokens:
            line_num = start[0]

            # check for line's length
            if end[1] > max_chars and not long_lines.has_key(line_num):
                long_lines[line_num] = 1
                msg = 'Too long line (max is %s)' % max_chars
                self.add_message('error', msg, line_num)

            if type == newline:
                # a program statement, or ENDMARKER, will eventually follow,
                # after some (possibly empty) run of tokens of the form
                #     (NL | COMMENT)* (INDENT | DEDENT+)?
                # If an INDENT appears, setting check_equal is wrong, and will
                # be undone when we see the INDENT.
                check_equal = 1

            elif type == indent:
                check_equal = 0
                try:
                    thisguy = indent_level(token, indent_string)
                    if not indents[-1] < thisguy:
                        msg = 'Indent not greater'
                        self.add_message('error', msg, line_num)
                    indents.append(thisguy)
                except IndentError, msg:
                    self.add_message('error', msg, line_num)

            elif type == dedent:
                # there's nothing we need to check here!  what's important is
                # that when the run of DEDENTs ends, the indentation of the
                # program statement (or ENDMARKER) that triggered the run is
                # equal to what's left at the top of the indents stack
                check_equal = 1
                if len(indents) > 1:
                    del indents[-1]

            elif check_equal and type not in junk:
                # this is the first "real token" following a NEWLINE, so it
                # must be the first token of the next program statement, or an
                # ENDMARKER; the "line" argument exposes the leading whitespace
                # for this statement; in the case of ENDMARKER, line is an empty
                # string, so will properly match the empty string with which the
                # "indents" stack was seeded
                if bad_construct is not None and bad_construct.match(line):
                    msg = 'Unauthorized construction'
                    self.add_message('warning', msg, line_num)
                check_equal = 0
                try:
                    thisguy = indent_level(line, indent_string)
                    if indents[-1] != thisguy:
                        msg = "Indent not equal"
                        self.add_message('error', msg, line_num)
                except IndentError, msg:
                    self.add_message('error', msg, line_num)
                    
        if line_num > self.config.max_module_lines:
            msg = 'Too many lines in module (%s)' % line_num
            self.add_message('error', msg)


def indent_level(string, indent):
    """ return the indent level of the string
    
    raise IndentError on ambiguous indentation
    """
    level = 0
    pos = 0
    unit_size = len(indent)
    while string[:unit_size] == indent:
        string = string[unit_size:]
        level += 1
        pos += unit_size
    if string and string[0] in ' \t':
        msg = 'column %s should not be [%s]' % (pos, string[0])
        raise IndentError(msg)
    return level


def register(linter):
    """ required method to auto register this checker """
    linter.register_checker(FormatChecker(linter))


